import json
import os
from sagemaker.s3 import S3Uploader, S3Downloader, s3_path_join

def upload(local_path, s3_location, kms_key=None):
    """
    Save a folder to an s3 location

    :param local_path: local path to upload
    :param s3_location: s3 location/folder to upload to
    :param kms_key: kms key to decrypt file
    :return:
    """
    return S3Uploader().upload(local_path, s3_location, kms_key=kms_key)

def download(s3_location, local_path, kms_key=None):
    """
    Download a folder from an s3 location

    :param s3_location: s3 location/folder to download
    :param local_path: local path to download to
    :param kms_key: kms key to decrypt file
    :return:
    """
    os.makedirs(local_path, exist_ok=True)
    return S3Downloader().download(s3_location, local_path, kms_key=kms_key)

def upload_config(config, s3_location, s3_file_name=None, kms_key=None):
    """
    Save a configuration document to an s3 location

    :param config: str/dict config json file path or configuration dict
    :param s3_location: s3 location/folder to upload config to
    :param s3_file_name: Name of the file in s3 folder. Required if config is a dict
    :param kms_key: kms key to decrypt file
    :return:
    """
    if type(config) == str:
        # assume it's a file path
        S3Uploader().upload(config, s3_location, kms_key=kms_key)
    else:
        assert type(config) == dict, "Configuration should be a local file or a python dictionary"
        assert s3_file_name is not None, "Configuration file name is required"
        config_str = json.dumps(config, indent=4)
        S3Uploader().upload_string_as_file_body(config_str, s3_path_join(s3_location, s3_file_name), kms_key=kms_key)


def get_config(s3_location, s3_file_name, kms_key=None, local_path=None):
    """
    Get a configuration document from s3

    :param s3_location: s3 location/folder to obtain config from
    :param s3_file_name: config file name at s3 location
    :param kms_key: kms key used to encrypt file
    :param local_path: local path to download configuration file to

    :return: configuration dict or None if local_path is passed
    """
    if local_path is not None:
        S3Downloader().download(s3_path_join(s3_location, s3_file_name), local_path, kms_key=kms_key)
    else:
        assert kms_key is None, "KMS encrypted files not supported for direct loading of configuration"
        json_str = S3Downloader().read_file(s3_path_join(s3_location, s3_file_name))
        return json.loads(json_str)


"""RGCN implementation adapted from https://github.com/dmlc/dgl/blob/master/examples/pytorch/rgcn-hetero/model.py
"""

from collections import defaultdict
from collections.abc import Mapping

import torch as th
import torch.nn as nn
import torch.nn.functional as F
import dgl
import dgl.nn as dglnn
import tqdm

class RelGraphConvLayer(nn.Module):
    r"""Relational graph convolution layer.

    Parameters
    ----------
    in_feat : int
        Input feature size.
    out_feat : int
        Output feature size.
    rel_names : list[str]
        Relation names.
    num_bases : int, optional
        Number of bases. If is none, use number of relations. Default: None.
    weight : bool, optional
        True if a linear layer is applied after message passing. Default: True
    bias : bool, optional
        True if bias is added. Default: True
    activation : callable, optional
        Activation function. Default: None
    self_loop : bool, optional
        True to include self loop message. Default: False
    dropout : float, optional
        Dropout rate. Default: 0.0
    """
    def __init__(self,
                 in_feat,
                 out_feat,
                 rel_names,
                 num_bases,
                 *,
                 weight=True,
                 bias=True,
                 activation=None,
                 self_loop=False,
                 dropout=0.0):
        super(RelGraphConvLayer, self).__init__()
        self.in_feat = in_feat
        self.out_feat = out_feat
        self.rel_names = rel_names
        self.num_bases = num_bases
        self.bias = bias
        self.activation = activation
        self.self_loop = self_loop

        self.conv = dglnn.HeteroGraphConv({
                str(rel) : dglnn.GraphConv(in_feat, out_feat, norm='right', weight=False, bias=False)
                for rel in rel_names
            })

        self.use_weight = weight
        self.use_basis = num_bases < len(self.rel_names) and weight
        if self.use_weight:
            if self.use_basis:
                self.basis = dglnn.WeightBasis((in_feat, out_feat), num_bases, len(self.rel_names))
            else:
                self.weight = nn.Parameter(th.Tensor(len(self.rel_names), in_feat, out_feat))
                nn.init.xavier_uniform_(self.weight, gain=nn.init.calculate_gain('relu'))

        # bias
        if bias:
            self.h_bias = nn.Parameter(th.Tensor(out_feat))
            nn.init.zeros_(self.h_bias)

        # weight for self loop
        if self.self_loop:
            self.loop_weight = nn.Parameter(th.Tensor(in_feat, out_feat))
            nn.init.xavier_uniform_(self.loop_weight,
                                    gain=nn.init.calculate_gain('relu'))

        self.dropout = nn.Dropout(dropout)

    def forward(self, g, inputs):
        """Forward computation

        Parameters
        ----------
        g : DGLHeteroGraph
            Input graph.
        inputs : dict[str, torch.Tensor]
            Node feature for each node type.

        Returns
        -------
        dict[str, torch.Tensor]
            New node features for each node type.
        """
        g = g.local_var()
        if self.use_weight:
            weight = self.basis() if self.use_basis else self.weight
            wdict = {self.rel_names[i] : {'weight' : w.squeeze(0)}
                     for i, w in enumerate(th.split(weight, 1, dim=0))}
        else:
            wdict = {}

        if g.is_block:
            inputs_src = inputs
            inputs_dst = {k: v[:g.number_of_dst_nodes(k)] for k, v in inputs.items()}
        else:
            inputs_src = inputs_dst = inputs

        hs = self.conv(g, inputs, mod_kwargs=wdict)

        def _apply(ntype, h):
            if self.self_loop:
                h = h + th.matmul(inputs_dst[ntype], self.loop_weight)
            if self.bias:
                h = h + self.h_bias
            if self.activation:
                h = self.activation(h)
            return self.dropout(h)
        return {ntype : _apply(ntype, h) for ntype, h in hs.items()}

class RGCNEncoder(nn.Module):
    r"""
    RGCN Encoder Module.

    Paper link: https://github.com/Jhy1993/Representation-Learning-on-Heterogeneous-Graph

    Parameters
    ----------
    etypes : list[str]
        list of edge types/relations in the graph
    in_size : int
        Size of input features
    hidden_size: int
        Size of output representations
    num_layers: int
        number of RGCN layers (Default: 1)
    num_bases: int
        number of bases for weight matrices (Default: -1 [use all etypes])
    use_self_loop: bool
        whether to use self loops (Default: True)
    dropout: float
        dropout rate for feature and attention weights (Default: 0.)
    activation: Callable
        callable activation function/layer or None, optional. (Default: torch.nn.Functional.relu)
    """

    def __init__(self,
                 etypes,
                 in_size,
                 hidden_size,
                 num_layers=1,
                 num_bases=-1,
                 use_self_loop=True,
                 dropout=0.,
                 activation=F.relu):
        super(RGCNEncoder, self).__init__()

        self.in_dim = in_size
        self.h_dim = hidden_size
        self.rel_names = sorted(etypes)
        if num_bases < 0 or num_bases > len(self.rel_names):
            self.num_bases = len(self.rel_names)
        else:
            self.num_bases = num_bases
        self.num_layers = num_layers
        self.dropout = dropout
        self.use_self_loop = use_self_loop
        self.layers = nn.ModuleList()

        self.layers.append(RelGraphConvLayer(self.in_dim, self.h_dim, self.rel_names,self.num_bases,
                                             activation=activation, self_loop=self.use_self_loop,
                                             dropout=self.dropout))

        for i in range(1, num_layers):
            self.layers.append(RelGraphConvLayer(self.h_dim, self.h_dim, self.rel_names, self.num_bases,
                                                 activation=activation, self_loop=self.use_self_loop,
                                                 dropout=self.dropout))

    def forward(self, g, h):
        for layer, block in zip(self.layers, g):
            h = layer(block, h)
        return h



    def batch_inference(self, g, x, batch_size, device, num_workers):
        """Minibatch inference of final representation over all node types.
        ***NOTE***
        For node classification, the model is trained to predict on only one node type's
        label.  Therefore, only that type's final representation is meaningful.

        Parameters
        ----------
        g : dgl.DGLGraph
            the graph to predict on
        x: dict[str: th.Tensor]
            the input features/embeddings for all nodes
        batch_size : int
            Number of target nodes in a single batch
        device: str
            Device to use for computation
        num_workers: int
            number of dataloader workers
        """

        for l, layer in enumerate(self.layers):
            y = {k: th.zeros(g.number_of_nodes(k),self.h_dim)
                for k in g.ntypes}

            sampler = dgl.dataloading.MultiLayerFullNeighborSampler(1)
            dataloader = dgl.dataloading.NodeDataLoader(
                g,
                {k: th.arange(g.number_of_nodes(k)) for k in g.ntypes},
                sampler,
                batch_size=batch_size,
                shuffle=True,
                drop_last=False,
                num_workers=num_workers)

            for input_nodes, output_nodes, blocks in tqdm.tqdm(dataloader):
                block = blocks[0].to(device)

                if not isinstance(input_nodes, Mapping):
                    input_nodes = {g.ntypes[0]: input_nodes}
                if not isinstance(output_nodes, Mapping):
                    output_nodes = {g.ntypes[0]: output_nodes}
                h = {k: x[k][input_nodes[k]].to(device) for k in input_nodes.keys()}
                h = layer(block, h)

                for k in h.keys():
                    y[k][output_nodes[k]] = h[k].cpu()

            x = y
        return y


import dgl
import torch as th
import torch.nn as nn

def initializer(emb):
    emb.uniform_(-1.0, 1.0)
    return emb

class NodeEmbedding(nn.Module):
    r"""
    Learnable node embeddings as initial node features

    Parameters
    ----------
    node_counts : dict[ntype: str, ncount: int]
        list of edge types/relations in the graph
    embed_size : int
        Size of embedding vector for each node
    use_torch: bool [Optional]
        whether to use torch embeddings or dgl Node Embeddings (Default: False)
    init_missing_embed_with: str [Optional]
        when loading from saved embeddings how to initialize embedding vectors for missing nodes (Default: 'random')
        options: ['random', 'average', 'zero']
    dev_id: int [Optional]
        device_id used in multigpu training (Default: 0)
    activation: Callable
        callable activation function/layer or None, optional. (Default: torch.nn.Functional.relu)
    """

    def __init__(self, node_counts, embed_size, use_torch=False, init_missing_embed_with='random', dev_id=0):
        super(NodeEmbedding, self).__init__()

        self.embed_size = embed_size
        self.use_torch_emb = use_torch
        self.init_missing_embed_with = init_missing_embed_with
        self.dev_id = dev_id
        self.node_embeds = nn.ModuleDict() if self.use_torch_emb else {}
        for ntype, ncount in node_counts.items():
            if self.use_torch_emb:
                self.node_embeds[str(ntype)] = nn.Embedding(ncount, self.embed_size)
            else:
                self.node_embeds[str(ntype)] = dgl.nn.NodeEmbedding(ncount, self.embed_size, name=str(ntype),
                                                                    init_func=initializer)


    def load_node_embeds(self, embs):
        for key, emb in embs.items():
            if self.use_torch_emb:
                if  self.node_embeds[key].weight.shape[0] == emb.shape[0]:
                    self.node_embeds[key] = nn.Embedding.from_pretrained(emb, freeze=True)
                else:
                    print("Initializing node embeddings for previously unseen nodes.")
                    if self.init_missing_embed_with == 'zero':
                        self.node_embeds[key].weight.data = nn.Parameter(th.zeros((self.node_embeds[key].weight.shape)))
                    elif self.init_missing_embed_with== 'average':
                        self.node_embeds[key].weight.data = nn.Parameter(th.mean(emb,dim=0,keepdim=True)
                                                                         .repeat(self.node_embeds[key].weight.shape[0],1))
                    self.node_embeds[key].weight.data[:emb.shape[0]] = nn.Parameter(emb[:])
            else:
                if self.node_embeds[key].emb_tensor.shape[0] == emb.shape[0]:
                    self.node_embeds[key].emb_tensor[:] = emb[:]
                else:
                    print("Initializing node embeddings for previously unseen nodes.")
                    if self.init_missing_embed_with == 'zero':
                        self.node_embeds[key].emb_tensor.data = nn.Parameter(th.zeros(self.node_embeds[key].emb_tensor.shape))
                    elif self.init_missing_embed_with == 'average':
                        self.node_embeds[key].emb_tensor.data = nn.Parameter(th.mean(emb,dim=0,keepdim=True).
                                                                        repeat(self.node_embeds[key].emb_tensor.shape[0],1))
                    self.node_embeds[key].emb_tensor.data[:emb.shape[0]]  = nn.Parameter(emb[:])

    def forward(self, node_ids):
        """Forward computation
        Parameters
        ----------
        node_ids : dict[ntype: str : ntype_nids: Tensor]
            node ids to generate embedding for.
        Returns
        -------
        tensor
            embeddings as the input of the next layer
        """

        embeds = {}
        for ntype in node_ids:
            if node_ids[str(ntype)].shape[0] == 0:
                continue
            if self.use_torch_emb:
                embeds[str(ntype)] = self.node_embeds[str(ntype)](node_ids[str(ntype)])
            else:
                embeds[str(ntype)] = self.node_embeds[str(ntype)](node_ids[str(ntype)], self.dev_id)
        return embeds

    @property
    def dgl_emb(self):
        """
        """
        if self.use_torch_emb:
            return None
        embs = [emb for emb in self.node_embeds.values()]
        return embs


import torch
import torch.nn as nn
import torch.nn.functional as F

import dgl.function as fn

from collections.abc import Sequence

class LabelPropagation(nn.Module):
    r"""
    Implementation of Label Propagation module
    Introduced in `Learning from Labeled and Unlabeled Data with Label Propagation <https://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.14.3864&rep=rep1&type=pdf>`_
    .. math::
        \mathbf{Y}^{\prime} = \alpha \cdot \mathbf{D}^{-1/2} \mathbf{A}
        \mathbf{D}^{-1/2} \mathbf{Y} + (1 - \alpha) \mathbf{Y},
    where unlabeled data is inferred by labeled data via propagation.
    Parameters
    ----------
        num_layers: int
            The number of propagations.
        alpha: float
            The :math:`\alpha` coefficient.
        adj_norm: str
            'DAD': D^-0.5 * A * D^-0.5
            'DA': D^-1 * A
            'AD': A * D^-1
    """

    def __init__(self, num_layers, alpha, adj_norm='DAD'):
        super(LabelPropagation, self).__init__()

        self.num_layers = num_layers
        self.alpha = alpha
        self.adj = adj_norm

    @torch.no_grad()
    def forward(self, g, labels, mask=None, clamp=True):
        with g.local_scope():
            if labels.dtype == torch.long:
                labels = F.one_hot(labels.view(-1)).to(torch.float32)

            y = labels
            if mask is not None:
                y = torch.zeros_like(labels)
                y[mask] = labels[mask]

            weighted_label = (1 - self.alpha) * y
            degs = g.in_degrees().float().clamp(min=1)
            norm = torch.pow(degs, -0.5 if self.adj == 'DAD' else -1).to(labels.device).unsqueeze(1)

            for _ in range(self.num_layers):
                # Assume the graphs to be undirected
                if self.adj in ['DAD', 'AD']:
                    y = norm * y

                g.ndata['h'] = y
                g.update_all(fn.copy_u('h', 'm'), fn.sum('m', 'h'))
                y = self.alpha * g.ndata.pop('h')

                if self.adj in ['DAD', 'DA']:
                    y = y * norm

                y = weighted_label + y
                if clamp:
                    y = y.clamp_(0., 1.)

        return y



from collections import defaultdict
from collections.abc import Mapping

import torch as th
import torch.nn as nn
import torch.nn.functional as F
import dgl
import dgl.nn as dglnn
import tqdm

class RelGATConvLayer(nn.Module):
    r"""Relational graph attention layer.

    Parameters
    ----------
    in_feat : int
        Input feature size.
    out_feat : int
        Output feature size.
    rel_names : list[str]
        Relation names.
    num_heads : int
        Number of attention heads.
    activation : callable, optional
        Activation function. Default: None
    self_loop : bool, optional
        True to include self loop message. Default: False
    dropout : float, optional
        Dropout rate. Default: 0.0
    """
    def __init__(self,
                 in_feat,
                 out_feat,
                 rel_names,
                 num_heads,
                 *,
                 activation=None,
                 dropout=0.0):
        super(RelGATConvLayer, self).__init__()
        self.in_feat = in_feat
        self.out_feat = out_feat
        self.rel_names = rel_names
        self.num_heads = num_heads
        self.activation = activation

        self.conv = dglnn.HeteroGraphConv({
                str(rel) : dglnn.GATConv(in_feat, out_feat, num_heads, allow_zero_in_degree=True)
                for rel in rel_names
            })

        self.dropout = nn.Dropout(dropout)

    def forward(self, g, inputs):
        """Forward computation

        Parameters
        ----------
        g : DGLHeteroGraph
            Input graph.
        inputs : dict[str, torch.Tensor]
            Node feature for each node type.

        Returns
        -------
        dict[str, torch.Tensor]
            New node features for each node type.
        """
        g = g.local_var()

        if g.is_block:
            inputs_src = inputs
            inputs_dst = {k: v[:g.number_of_dst_nodes(k)] for k, v in inputs.items()}
        else:
            inputs_src = inputs_dst = inputs

        hs = self.conv(g, inputs)

        def _apply(ntype, h):
            if self.activation:
                h = self.activation(h.view(-1, self.out_feat * self.num_heads))
            return self.dropout(h)
        return {ntype : _apply(ntype, h) for ntype, h in hs.items()}


class RGATEncoder(nn.Module):
    r"""
    RGAT Encoder Module. Extending GAT to heterogeneous graphs

    Parameters
    ----------
    etypes : list[str]
        list of edge types/relations in the graph
    in_size : int
        Size of input features
    hidden_size: int
        Size of output representations
    num_layers: int
        number of RGCN layers (Default: 1)
    num_heads: int
        number of bases for weight matrices (Default: 1)
    dropout: float
        dropout rate for feature and attention weights (Default: 0.)
    activation: Callable
        callable activation function/layer or None, optional. (Default: torch.nn.Functional.relu)
    """

    def __init__(self,
                 etypes,
                 in_size,
                 hidden_size,
                 num_layers=1,
                 num_heads=1,
                 dropout=0.,
                 activation=F.elu):
        super(RGATEncoder, self).__init__()

        self.in_dim = in_size
        self.h_dim = hidden_size
        self.rel_names = sorted(etypes)
        self.num_heads = num_heads
        self.num_layers = num_layers
        self.dropout = dropout
        self.layers = nn.ModuleList()

        self.layers.append(RelGATConvLayer(self.in_dim, self.h_dim, self.rel_names,self.num_heads,
                                             activation=activation, dropout=self.dropout))

        for i in range(1, num_layers):
            self.layers.append(RelGATConvLayer(self.h_dim * self.num_heads, self.h_dim, self.rel_names, self.num_heads,
                                                 activation=activation, dropout=self.dropout))

    def forward(self, g, h):
        for layer, block in zip(self.layers, g):
            h = layer(block, h)
        return h



    def batch_inference(self, g, x, batch_size, device, num_workers):
        """Minibatch inference of final representation over all node types.
        ***NOTE***
        For node classification, the model is trained to predict on only one node type's
        label.  Therefore, only that type's final representation is meaningful.

        Parameters
        ----------
        g : dgl.DGLGraph
            the graph to predict on
        x: dict[str: th.Tensor]
            the input features/embeddings for all nodes
        batch_size : int
            Number of target nodes in a single batch
        device: str
            Device to use for computation
        num_workers: int
            number of dataloader workers
        """

        for l, layer in enumerate(self.layers):
            y = {k: th.zeros(g.number_of_nodes(k),self.h_dim*self.num_heads)
                for k in g.ntypes}

            sampler = dgl.dataloading.MultiLayerFullNeighborSampler(1)
            dataloader = dgl.dataloading.NodeDataLoader(
                g,
                {k: th.arange(g.number_of_nodes(k)) for k in g.ntypes},
                sampler,
                batch_size=batch_size,
                shuffle=True,
                drop_last=False,
                num_workers=num_workers)

            for input_nodes, output_nodes, blocks in tqdm.tqdm(dataloader):
                block = blocks[0].to(device)

                if not isinstance(input_nodes, Mapping):
                    input_nodes = {g.ntypes[0]: input_nodes}
                if not isinstance(output_nodes, Mapping):
                    output_nodes = {g.ntypes[0]: output_nodes}
                h = {k: x[k][input_nodes[k]].to(device) for k in input_nodes.keys()}
                h = layer(block, h)

                for k in h.keys():
                    y[k][output_nodes[k]] = h[k].cpu()

            x = y
        return y


import boto3
from datetime import datetime
import time

from .cloudwatch_utils import print_logs

try:
    boto_session = boto3.DEFAULT_SESSION or boto3.Session()
    region_name = boto_session.region_name
    batch = boto3.client(
        service_name='batch',
        region_name=region_name)
except:
    print("Could not initialize boto client")

batch_log_group_name = '/aws/batch/job'

def wait_for_job(jobId):
    spin = ['-', '/', '|', '\\', '-', '/', '|', '\\']
    spinner = 0
    running = False
    start_time = 0
    wait = True

    while wait:
        time.sleep(1)
        describeJobsResponse = batch.describe_jobs(jobs=[jobId])
        status = describeJobsResponse['jobs'][0]['status']
        if status == 'SUCCEEDED' or status == 'FAILED':
            print('%s' % ('=' * 80))
            print('Job [%s] %s' % (jobId, status))
            break
        elif status == 'RUNNING':
            log_stream_name = describeJobsResponse['jobs'][0]['container']['logStreamName']
            if not running and log_stream_name:
                running = True
                print('\rJob [%s] is RUNNING.' % (jobId), flush=True, end='')
                print('\nOutput [%s]:\n%s' % (log_stream_name, '=' * 80))
            if log_stream_name:
                start_time = print_logs(batch_log_group_name, log_stream_name, start_time) + 1
        else:
            print('\rJob [%s] is %-9s... %s' % (jobId, status, spin[spinner % len(spin)]), flush=True, end='')
            spinner += 1


import os
import json
import torch as th
from collections.abc import Mapping
from ..utils import get_available_devices

def get_training_config():
    r"""

    Returns
    -------
    tuple(data_path: str, model_path: str, devices: list[int])
    """
    data_path = str(os.environ['SM_CHANNEL_TRAIN'])
    model_path = '/opt/ml/checkpoints'
    devices = get_available_devices()

    return data_path, model_path, devices

def get_train_nids(g, target_ntype=None):
    r"""

    get validation node ids from the graph

    Parameters
    ----------
    g : dgl.DGLGraph
        the Graph
    target_ntype : str
        the node type of the targets
    Returns
    -------
    train_idx: th.Tensor
    """
    train_mask = g.ndata['train_mask'] if target_ntype is None else g.nodes[target_ntype].data['train_mask']
    train_idx = th.arange(train_mask.shape[0])
    train_idx = train_idx[train_mask.bool()]
    return train_idx

def get_valid_nids(g, target_ntype=None):
    r"""

    get validation node ids from the graph

    Parameters
    ----------
    g : dgl.DGLGraph
        the Graph
    target_ntype : str
        the node type of the targets
    Returns
    -------
    valid_idx: th.Tensor
    """
    valid_mask = g.ndata['valid_mask'].bool() if target_ntype is None else g.nodes[target_ntype].data['valid_mask']
    valid_idx = th.arange(valid_mask.shape[0])
    valid_idx = valid_idx[valid_mask.bool()]
    return valid_idx

def get_node_labels(g, target_ntype=None):
    r"""

    get node labels from the graph

    Parameters
    ----------
    g : dgl.DGLGraph
        the Graph
    target_ntype : str (optional)
        the node type of the targets
   Returns
   -------
   labels: th.Tensor
   """
    labels = g.ndata['labels'] if target_ntype is None else g.nodes[target_ntype].data['labels']
    return labels

def get_train_eids(g, target_etype=None, include_reverse_etypes=True):
    r"""

    get validation node ids from the graph

    Parameters
    ----------
    g : dgl.DGLGraph
        the Graph
    target_etype : str
        the edge type of the targets
    include_reverse_etypes : bool (default: True)
        whether to include edge ids from reversed edge types.
    Returns
    -------
    train_idx: th.Tensor
    """
    train_mask = {target_etype: g.edges[target_etype].data['train_mask']} if target_etype  else g.edata['train_mask']
    if include_reverse_etypes and isinstance(train_mask, Mapping):
        for etype in train_mask:
            src_type, rel_type, dst_type = etype
            if "rev-" in rel_type and (dst_type, rel_type.replace("rev-", ""), src_type) in train_mask:
                train_mask[etype] = g.edata['rev_train_mask'][etype]
    if isinstance(train_mask, Mapping):
        train_idx = {}
        for etype in train_mask:
            idx = th.arange(train_mask[etype].shape[0])
            train_idx[etype] = idx[train_mask[etype].bool()]
    else:
        train_idx = th.arange(train_mask.shape[0])
        train_idx = train_idx[train_mask.bool()]
    return train_idx

def get_valid_eids(g, target_etype=None, include_reverse_etypes=False):
    r"""

    get validation node ids from the graph

    Parameters
    ----------
    g : dgl.DGLGraph
        the Graph
    target_etype : str
        the node type of the targets
    include_reverse_etypes : bool (default: False)
        whether to include edge ids from reversed edge types.
    Returns
    -------
    valid_idx: th.Tensor or dict[etype:str : th.Tensor]
    """
    valid_mask = {target_etype: g.edges[target_etype].data['valid_mask']} if target_etype  else g.edata['valid_mask']
    if include_reverse_etypes and isinstance(train_mask, Mapping):
        for etype in valid_mask:
            src_type, rel_type, dst_type = etype
            if "rev-" in rel_type and (dst_type, rel_type.replace("rev-", ""), src_type) in valid_mask:
                valid_mask[etype] = g.edata['rev_valid_mask'][etype]
    if isinstance(valid_mask, Mapping):
        valid_idx = {}
        for etype in valid_mask:
            idx = th.arange(valid_mask[etype].shape[0])
            valid_idx[etype] = idx[valid_mask[etype].bool()]
    else:
        valid_idx = th.arange(valid_mask.shape[0])
        valid_idx = valid_idx[valid_mask.bool()]
    return valid_idx

def get_edge_labels(g, target_etype=None):
    r"""

    get node labels from the graph

    Parameters
    ----------
    g : dgl.DGLGraph
        the Graph
    target_etype : str (optional)
        the edge type of the targets
   Returns
   -------
   labels: th.Tensor
   """
    labels = g.edata['labels'] if target_etype is None else g.edges[target_etype].data['labels']
    return labels

def get_all_forward_edges(g):
    r"""

    get all forward (non-reversed) edges from the graph in the u, v, etype_id

    Parameters
    ----------
    g : dgl.DGLGraph
        the Graph

   Returns
   -------
   labels: (u: th.Tensor, v: th.Tensor, etype_ids: th.Tensor)
   """
    srcs, dsts, etype_ids = [], [], []
    for etype_id, etype in enumerate(g.canonical_etypes):
        if "rev-" in etype[1]:
            continue
        u, v = g.edges(etype=etype)
        srcs.append(u)
        dsts.append(v)
        etype_ids.append(th.full_like(u, etype_id))

    return th.cat(srcs), th.cat(dsts), th.cat(etype_ids)



import os
import dgl
import pickle

class GraphLoader(object):
    r""" Load DGLGraph features and metadata created by Neptune ML DataProcessing Job.

    Parameters
    ----------
    path : str
        path where graph binary and metadata are stored
    name : str, optional
        name of the graph loader

    >>> graph_loader = neptuneml.GraphLoader()

    ** Get the graph **
    >>> g = graphloader.graph

    ** Get the node features **
     >>> g = graphloader.get_node_features()

    ** Get the edge features **
     >>> g = graphloader.get_edge_features()
    """
    def __init__(self, path, name='graph'):
        self._name = name
        self._graph = None
        self._node_dict = {}
        self._label_map = None
        self._is_multilabel = False
        self._label_properties = None
        self._node_feat_meta = None
        self._edge_feat_meta = None
        self.load(path)

    def load(self, path):
        """load the graph and the labels

        Parameters
        ----------
        path: str
            Path where to load the graph and the labels
        """
        graph_path = os.path.join(path, 'graph.bin')
        info_path = os.path.join(path, 'info.pkl')
        graphs, _ = dgl.data.load_graphs(graph_path)
        self._g = graphs[0]

        with open(info_path, "rb") as pf:
            info = pickle.load(pf)
        self._node_dict = info['node_id_map']
        self._label_map = info['label_map']
        self._label_properties = info['label_properties']
        self._is_multilabel = info['is_multilabel']
        self._node_feat_meta = info['node_feat_meta']
        self._edge_feat_meta = info['edge_feat_meta']

    def get_node_features(self):
        assert self._node_feat_meta is not None and self._g is not None,\
            "Graph has not been loaded. you must load graph before getting node features. run `graphloader.load()`"
        nf = {ntype: {} for ntype in self._node_feat_meta}
        for ntype in self._node_feat_meta:
            for feature in self._node_feat_meta[ntype]:
                if self._node_feat_meta[ntype][feature]['type'] == 'raw_feat':
                    nf[ntype][feature] = self._node_feat_meta[ntype][feature]['raw_feat']
                else:
                    nf[ntype][feature] = self._g.nodes[ntype].data[feature]
        return nf

    def get_edge_features(self):
        assert self._edge_feat_meta is not None and self._g is not None, \
            "Graph has not been loaded. you must load graph before getting edge features. run `graphloader.load()`"
        ef = {etype: {} for etype in self._edge_feat_meta}
        for etype in self._edge_feat_meta:
            for feature in self._edge_feat_meta[etype]:
                if self._edge_feat_meta[etype][feature]['type'] == 'raw_feat':
                    ef[etype][feature] = self._edge_feat_meta[etype][feature]['raw_feat']
                else:
                    ef[etype][feature] = self._g.edges[etype].data[feature]
        return ef

    @property
    def rtype2id(self):
        """ Return mappings from relation type to internal node id

        Note: used only for KGE models

        Return
        ------
        dict of dict:
            {relation_type : {raw relation id(string/int): dgl_id}}
        """
        return self._rel_dict

    @property
    def node2id(self):
        """ Return mappings from raw node id/name to internal node id

        Return
        ------
        dict of dict:
            {node_type : {raw node id(string/int): dgl_id}}
        """
        return self._node_dict

    @property
    def id2node(self):
        """ Return mappings from internal node id to raw node id/name

        Return
        ------
        dict of dict:
            {node_type : {raw node id(string/int): dgl_id}}
        """
        return {node_type: {val: key for key, val in node_maps.items()} \
                for node_type, node_maps in self._node_dict.items()}

    @property
    def label_map(self):
        """ Return mapping from internal label id to original label

        Return
        ------
        dict:
            {type: {label id(int) : raw label(string/int)}}
        """
        return self._label_map

    @property
    def label_properties(self):
        """ Return mapping from node/edge type to label column name or property

        Return
        ------
        dict:
            {type: label_property (string/int)}
        """
        return self._label_properties

    @property
    def graph(self):
        """ Return graph
        """
        return self._g

    @property
    def is_multilabel(self):
        """ Return whether label is multilabel or singlelebel
        """
        return self._is_multilabel

    @property
    def edge_feat_meta(self):
        return self._edge_feat_meta

    @property
    def node_feat_meta(self):
        return self._node_feat_meta



#!/usr/bin/env python3

# -*- coding: utf-8 -*-
#
# setup.py
#
# Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from setuptools import find_packages
from setuptools import setup

required_packages = [
    "boto3>=1.26.28,<2.0",
    "numpy>=1.9.0,<2.0",
    "graph-notebook==3.7.0",
    "scikit-learn>=0.24.1",
    "sagemaker>=2.48.1",
]

VERSION = '0.0.2.dev4'
setup(
    name='neptuneml-toolkit',
    version=VERSION,
    description='Open source library for training models with Amazon Neptune ML',
    maintainer='Neptune ML Team',
    maintainer_email='sojiadeshina@gmail.com',
    packages=find_packages(),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: Apache Software License',
    ],
    license='Apache 2.0',
    install_requires=required_packages,
)



import torch as th
import torch.nn as nn
import dgl.function as fn

class DotProductDecoder(nn.Module):
    r"""
    Implementation of DotProduct Decoder.

    Suitable when predicting a logit for each edge where
    edge is represented by a pair of arrays of node features

    .. math::

        f(u,v) = \textbf{u}^{T}\textbf{v} = \sum_{i=1}^{d}\textbf{u}_i \cdot \textbf{v}_i

    """
    def __init__(self):
        super(DotProductDecoder, self).__init__()

    def forward(self, u, v):
        """Forward function.
        Compute logits for each pair ``(ufeat[i], vfeat[i])``.
        Parameters
        ----------
        u : th.Tensor
            Source nodes embeddings. Shape: (B, D)
        v : th.Tensor
            Destination nodes embeddings. Shape: (B, D)
        etype_ids: th.Tensor
            Edge type ids for each edge represented by u[i], v[i]
        Returns
        -------
        th.Tensor
            Representation for edge between u and v. Shape: (B, out)
        """
        assert u.shape == v.shape, "Shape mismatch between u and v"
        score = th.einsum('bi,bi->b', u, v)
        return score

class GraphDotProductDecoder(nn.Module):
    r"""
    Implementation of DotProduct Decoder using DGL Message Passing API.

    Suitable when predicting a logit for each edge where
    edge is represented by a pair of arrays of node features

    .. math::

        f(u,v) = \textbf{u}^{T}\textbf{v} = \sum_{i=1}^{d}\textbf{u}_i \cdot \textbf{v}_i

    """

    def __init__(self):
        super(GraphDotProductDecoder, self).__init__()

    def forward(self, graph, h):
        """Forward function.
        Compute output for each edge ``(u, v)`` in the graph using `h_u` and h_v`.
        Parameters
        ----------
        graph : dgl.Graph or dgl.Block
            The graph structure with the same number of nodes as h
        h : th.Tensor or dict[ntype: th.Tensor]
            Nodes embeddings. Shape: (N, D)
        Returns
        -------
        th.Tensor
            Representation for edge between u and v. Shape: (B, out)
        """
        with graph.local_scope():
            graph.ndata['h'] = h
            for etype in graph.canonical_etypes:
                graph.apply_edges(fn.u_dot_v('h', 'h', 'out'), etype=etype)
            return graph.edata['out']